/*      8xxtarg.c   Revision 2.3  1/24/97                             */

/* This software was written by Symbios Logic Inc. to develop and test new
 * products.  Symbios Logic assumes no liability for its use.  This software
 * is released to the public domain to illustrate certain programming
 * techniques for the 53c8xx chips in target mode.
 *
 */

/*
 * Revision 2.3 1/24/97  Scott Rawlings
 * Added support for the 53C895.  Fixed major bug related to working with
 * SCRIPTS in SCRIPTS RAM that started out mis-aligned in memory.  Added
 * support for the clock doubler in newer revisions of the 875.
 *
 * Revision 2.22 4/9/96  Scott Rawlings
 * Added capability to use physical memory greater than 64 Meg.
 *
 * Revision 2.21 4/2/96  Scott Rawlings
 * Fixed problem with wide negotiation when the initiator has previously
 * negotiated wide and now wants to change to narrow.  Added ability to use a
 * cache size of 512 bytes.  Fixed problem with inquiry command when the
 * initiator does synchronous and/or wide negotiation during the message out
 * phase.
 *
 * Revision 2.2 2/12/96  Scott Rawlings
 * Changed the save and restore RAM drive to use fread and fwrite commands for
 * speed inprovement.  Added code to long word align the data transfer buffer
 * at run time.  Fixed major problem with disconnect logic on write commands.
 * Changed synchronous setup method from calculation on the fly to table
 * lookup.  This allows the code to select the best period match rather than
 * the first close match.  Fixed menu display problem encountered when
 * synchronous and wide negotiation happen multiple times without a reset.
 * Added user notification of SCSI bus reset.  Added ability for user to alter
 * SCSI ID without restarting program.  Consolidated some redundant code into
 * function calls.
 *
 * Revision 2.11 11/14/95  Scott Rawlings
 * Added support for differential boards.
 *
 * Revision 2.1 11/13/95  Scott Rawlings
 * Added capability to change the drive cache size without having to re-start
 * the program.  Removed some redundant code and cleaned up some inefficient
 * code.
 *
 * Revision 2.0 10/31/95  Scott Rawlings
 * Changed from ISTAT polling to IRQ driven control.  Removed all EMM calls and
 * changed to direct use of upper memory area for faster operation.
 * Streamlined code for better performance.  Added menu options for more user
 * control and ease of use.  Removed most inline assembly code.
 *
 * Revision 1.0 9/1/95  Scott Rawlings
 * Initial release.
 */

#include <dos.h>
#include <bios.h>
#include <mem.h>
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <alloc.h>
#include <math.h>

#include "8xxtarg.out" /* contains the code for the SCSI SCRIPTS */

#include "types.h"
#include "tools.h"
#include "table.h"
#include "tools_lm.c"
#include "ioport32.h"
#include "pci.h"
#include "pci_8xx.h"

#define PIC2_MASK_REG 0xA1 /* slave interrupt cont. mask reg */
#define PIC2_EOI_REG 0xA0 /* end of interrupt register, slave */
#define PIC1_EOI_REG 0x20 /* end of interrupt register, master */
#define IRQ9 0x71	/* vector of IRQ9 */
#define IRQ10 0x72 /* vector of IRQ10 */
#define IRQ11 0x73 /* vector of IRQ11 */
#define IRQ12 0x74 /* vector of IRQ12 */
#define IRQ13 0x75 /* vector of IRQ13 */
#define IRQ14 0x76 /* vector of IRQ14 */
#define IRQ15 0x77 /* vector 0f IRQ15 */
#define IRQ9_mask 0xFD /* mask for enabling IRQ9 */
#define IRQ10_mask 0xFB	/* mask for enabling IRQ10 */
#define IRQ11_mask 0xF7	/* mask for enabling IRQ11 */
#define IRQ12_mask 0xEF	/* mask for enabling IRQ12 */
#define IRQ13_mask 0xDF	/* mask for enabling IRQ13 */
#define IRQ14_mask 0xBF	/* mask for enabling IRQ14 */
#define IRQ15_mask 0x7F	/* mask for enabling IRQ15 */

/* Prototypes */
static void	init_siop(void);
static void init_table(void);
static void init_script(void);
static void init_irq(void);
static void release_irq(void);
static void	__interrupt isr(void);
static void print_menu(void);
static void device_reset(void);

static char version[] = "8xxtarg.exe  Version 2.3  1/24/97  Scott Rawlings";

device_record device[4]; /* structure for all 8xx devices found on PCI bus */
int selected_device; /* the number of the device that the user selected */
ubyte old_irq_mask; /* storage of old mask for restore at quit time */
uquad *script; /* pointer to SCSI SCRIPTS */
table *buffer_table, *dummy_table; /* the tables used for table-indirect */

/* buffer for memory moves.  the difference between the declared size and */
/* the size variable is so that the buffer can be aligned to a long word */
/* boundary if it doesn't start there */
ubyte data_transfer_buf[32771U];
ubyte *ptr_data_transfer_buf;
#define	data_transfer_buf_size	32768U

/* 8-bit registers */
ubyte dstat, sist0, sist1, istat, scntl3, ssid, scid;
ubyte sxfer, dcmd, ctest2, ctest3;

/* 16-bit registers */
uword respid;

/* 32-bit registers */
uquad dsp, dsa, dcmd_dbc, dbc, vector, addr, dsps;

/* current SCSI phase */
ubyte phase;

/* other global variables */
int clk, id, wide_chip = 0, max_offset = 8, done, last_synch_entry;
int i, p, x, first = 1, max_cd, period, xferp, offset, error, busy = 0;
int return_code, return_flag = 2, got_int, done_synch = 0, done_wide = 0;
int buffer_off = 0, doubler = 0, quadrupler = 0;
unsigned int block_size, k, m;
unsigned long int j, data_buf_size, bytes_left, byte_count, lba, size;
float tp, cd[7] = {1, 1.5, 2, 3, 4, 6, 8};

/* structure that contains a synchronous period, and the clock dividor and */
/* xferp value to create that period */
struct synch_table
{
 int synch_period, xferp_value;
 float clock_dividor;
 ubyte synch_answer;
};

/* array that will hold all possible synchronous periods that can be made */
/* with the current SCSI clock, plus a temporary holder for the sort */
struct synch_table synch_lookup_table[56], temp_holder;

/* chip base addresses and 32-bit register addresses */
uword io_port;
uquad RAM_address = 0;
uquad scripts_address;
uword dsp_address, dsa_address, dsps_address;

/* SCRIPTS entry points */
uquad wait_select, msg_out_phase, command_phase, tur, request_sense;
uquad read_return, read_reconnect, write_return, write_reconnect;
uquad msg_in_phase, inquiry, read_capacity, copy_data, synch_wide_neg_return;
uquad stopped_busy_wait_select, stopped_busy_tur;

/* other buffers */
ubyte command_buf[12];
ubyte message_out_buf[1];
ubyte dummy_command_buf[12];
ubyte dummy_message_out_buf[1];
ubyte sync_neg[2];
ubyte wide_neg[1];

/* pre-set buffers */
static ubyte message_reject[] =
 {
  0x07
 };

static ubyte identify_msg[] =
 {
  0x80
 };

static ubyte command_complete[] =
 {
  0x00
 };

static ubyte device_busy[] =
 {
  0x08
 };

static ubyte status_good[] =
 {
  0x00
 };

static ubyte check_condition[] =
 {
  0x02
 };

static ubyte save_pointers_msg[] =
 {
  0x02
 };

static ubyte restore_pointers_msg[] =
 {
  0x03
 };

static ubyte sel_info[] =
 {
  0x00, 0x00, 0x07, 0x33
 };

 static ubyte synch_msg_in_buf[] =
 {
  0x01, 0x03, 0x01, 0x19, 0x08
 };

static ubyte wide_msg_in_buf[] =
 {
  0x01, 0x02, 0x03, 0x00
 };

static ubyte disc_msg[] =
 {
  0x02, 0x04
 };

static ubyte read_capacity_data[] =
 {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
 };

static ubyte inquiry_data[] =
 {
  0x00, 0x00, 0x02, 0x02, 0x1f, 0x00, 0x00, 0x10, 0x53, 0x59, 0x4d, 0x42,
  0x49, 0x4f, 0x53, 0x20, 0x52, 0x41, 0x4d, 0x20, 0x44, 0x49, 0x53, 0x4b,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x31, 0x2e, 0x30, 0x30
 };

static ubyte bad_lun_inquiry_data[] =
 {
  0x7f, 0x00, 0x02, 0x02, 0x1f, 0x00, 0x00, 0x10, 0x53, 0x59, 0x4d, 0x42,
  0x49, 0x4f, 0x53, 0x20, 0x52, 0x41, 0x4d, 0x20, 0x44, 0x49, 0x53, 0x4b,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x31, 0x2e, 0x30, 0x30
 };

static ubyte no_sense_data[] =
 {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00
 };

static ubyte reset_sense_data[] =
 {
  0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x00, 0x00, 0x00, 0x00,
  0x29, 0x00, 0x00, 0x00, 0x00, 0x00
 };

static ubyte unit_not_ready[] =
 {
  0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x00, 0x00, 0x00, 0x00,
  0x04, 0x00, 0x00, 0x00, 0x00, 0x00
 };

static ubyte illegal_command[] =
 {
  0x00, 0x00, 0x05, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x00, 0x00, 0x00, 0x00,
  0x24, 0x00, 0x00, 0x00, 0x00, 0x00
 };

main()
{
 /* local only variables */
 int max_devices = 4, num_devices;
 unsigned int new_block_size, value, inc_value, lba_offset;
 unsigned int reg, dword_offset, max_id = 7, mem_size;
 unsigned long int file_size, use_lba, last_lba, config_value, mask_value;
 unsigned long int memory;
 char file_name[32], cont;
 device_record *device_ptr[4];
 FILE *save_file;

 clrscr();
 printf("%s\n\n", version);

 /* check the data_transfer_buf alignment and fix it if needed */
 if (FP_OFF(data_transfer_buf) & 0x01)
  ptr_data_transfer_buf = &data_transfer_buf[0] + 3;
 else
  if (FP_OFF(data_transfer_buf) & 0x02)
	ptr_data_transfer_buf = &data_transfer_buf[0] + 2;
  else
	if (FP_OFF(data_transfer_buf) & 0x03)
	 ptr_data_transfer_buf = &data_transfer_buf[0] + 1;
	else
	 ptr_data_transfer_buf = &data_transfer_buf[0];

 /* initial startup values need to be gotten from user */
 do
  {
	printf("Bytes Per Logical Block (512, 1024, 2048, 4096) = ?");
	scanf("%lu", &block_size);
  }while ((block_size != 512) && (block_size != 1024)    /* block size is */
		 && (block_size != 2048) && (block_size != 4096)); /* limited to these */
																			/* values */

 asm	mov	ah, 88h /* function 88h of int 15h returns in ax how many */
 asm	int	15h     /* contiguous 1k blocks exist above 1 Meg */
 asm	mov	mem_size, ax

 /* if there is more than 63MBytes of extended memory, the int 15 call can't */
 /* see it.  in this case, we must have the user input the amount of system */
 /* memory that exists */
 if (mem_size >= 0xfc00)
  {
	printf("You have more than 64MBytes in the system.\n");
	printf("Please input the total amount of system RAM in MBytes: ");
	scanf("%lu", &memory);
	memory -= 1;  /* 1 meg for system ram */
	memory *= 1024;  /* translate value to KBytes */
  }
 else
  memory = (unsigned long int)mem_size; /* value already in KBytes */

 last_lba = (memory * 1024) / block_size; /* calculate the last lba in mem */
 last_lba -= 1; /* back off by one logical block, just to be safe */
 size = last_lba * block_size; /* size in bytes */
 last_lba -= 1; /* the last lba is one less because we start at zero */

 /* display the drive size for the user */
 printf("Drive Size = %.2f MBytes (%lu Bytes)\n",
															(((float)size) / 1048576L), size);

 /* set up data for read capacity commands */
 for (i = 0, x = 3; i < 4; i++, x--)
  read_capacity_data[i] = (int)((last_lba >> 8 * x) & 0xff);
 for (i = 0, x = 3; i < 4; i++, x--)
  read_capacity_data[i + 4] = (block_size >> 8 * x) & 0xff;

 /* this buffer size determines how much data can be sent before a disconnect */
 /* or save pointers message occurs */
 do
  {
	printf("Size Of Drive Cache (In Kilobytes <= 1024K, 0 = 512 bytes) = ?");
	scanf("%u", &data_buf_size);
  }while (data_buf_size > 1024);

 /* a value of zero gives a cache size of 512 bytes */
 if (data_buf_size == 0)
  data_buf_size = 512;
 else
  data_buf_size *= 1024;

 /* these functions find all Symbios devices on the PCI bus and ask which */
 /* device should be used by this code */
 for (i = 0; i <= max_devices ; i++)
  device_ptr[i] = &device[i];
 num_devices = C8XX_BuildDeviceList(&device_ptr, max_devices);
 selected_device = C8XX_ChooseDevice(&device_ptr, num_devices);

 /* get the IO base address of the chosen chip */
 error = C8XX_GetIOBase(&device[selected_device]);
 if (error)
  {
	printf("Selected Chip Has No I/O Address\n");
	exit(1);
  }
 io_port = (uword)(device[selected_device].base_addr2 & 0xfffe);

 printf("SCLK Speed In MHz = ?"); /* get the SCLK value from the user */
 scanf("%d", &clk);

 /* if this is a newer 875, or an 895, and the clock speed is between 40Mhz */
 /* and 50Mhz, then we must ask the user if the clock doubler or quadrupler */
 /* should be used */
 if ((device[selected_device].device_id == C875_DEVICE_ID) &&
	 ((device[selected_device].rev_id & 0x0f) >= 0x02) &&
	  (clk <= 50))
  do
	{
	 printf("Clock Doubler (0 = Off, 1 = On) = ?");
	 scanf("%d", &doubler);
	 if (doubler)
	  clk *= 2;
	}while ((doubler != 0) && (doubler != 1));

 /* clock quadrupler can only work with 40MHz */
 if ((device[selected_device].device_id == C895_DEVICE_ID) && (clk == 40))
  do
	{
	 printf("Clock Quadrupler (0 = Off, 1 = On) = ?");
	 scanf("%d", &quadrupler);
	 if (quadrupler)
	  clk *= 4;
	}while ((quadrupler != 0) && (quadrupler != 1));

 tp = ((1 / (float)clk) * 1000); /* clock period */

 /* set up the wide chip flag and the inquiry data if the chip supports wide */
 if ((device[selected_device].device_id == C875_DEVICE_ID)
	| (device[selected_device].device_id == C825_DEVICE_ID)
	| (device[selected_device].device_id == C895_DEVICE_ID))
  {
	do
	 {

	  /* if chip is wide, ask user if cable is also wide */
	  printf("SCSI Bus Width (0 = Narrow, 1 = Wide) = ?");
	  scanf("%d", &wide_chip);
	 } while((wide_chip != 0) && (wide_chip != 1));

	/* if cable is wide, set up inquiry data to say we can do wide */
	if (wide_chip)
	 {
	  inquiry_data[7] = 0x30;
	  bad_lun_inquiry_data[7] = 0x30;
	  max_id = 15;
	 }
  }

 /* get SCSI id from user */
 do
  {
	printf("SCSI ID = ?");
	scanf("%u", &id);
  }while (id > max_id);

 init_siop(); /* reset chip and set up initial operating values */
 init_table(); /* malloc memory and set up table */
 init_script(); /* make sure SCRIPTS start on 32-bit boundary */

 /* calculate all possible periods with each clock dividor and xferp and put */
 /*them in the lookup table */
 i = 0;
 for (p = 0; p < max_cd; p++)
  for (xferp = 0; xferp <= 7; xferp++, i++)
	{
	 synch_lookup_table[i].xferp_value = xferp;
	 synch_lookup_table[i].clock_dividor = cd[p];
	 period = (int)((tp * cd[p]) * (xferp + 4));
	 synch_lookup_table[i].synch_period = period;

	 /* new values were decided upon by the SCSI commitee for Ultra/Ultra2 */
	 /* these values don't follow the previous convention, and must be treated */
	 /* as special cases */
	 if ((period >= 25) && (period < 30))
	  synch_lookup_table[i].synch_answer = 0x0a;
	 else
	  if ((period >=30) && (period < 50))
		synch_lookup_table[i].synch_answer = 0x0b;
	  else
		synch_lookup_table[i].synch_answer = (ubyte)(period / 4);
	}
 last_synch_entry = i - 1;

 /* order all of the synch periods from smallest to largest */
 for (i = last_synch_entry; i >= 0; i--)
  for (p = 1; p <= i; p++)
	if (synch_lookup_table[p - 1].synch_period >
														  synch_lookup_table[p].synch_period)
	 {
	  temp_holder.synch_period = synch_lookup_table[p - 1].synch_period;
	  temp_holder.xferp_value = synch_lookup_table[p - 1].xferp_value;
	  temp_holder.clock_dividor = synch_lookup_table[p - 1].clock_dividor;
	  temp_holder.synch_answer = synch_lookup_table[p - 1].synch_answer;

	  synch_lookup_table[p - 1].synch_period =
														  synch_lookup_table[p].synch_period;
	  synch_lookup_table[p - 1].xferp_value = synch_lookup_table[p].xferp_value;
	  synch_lookup_table[p - 1].clock_dividor =
														 synch_lookup_table[p].clock_dividor;
	  synch_lookup_table[p - 1].synch_answer =
														  synch_lookup_table[p].synch_answer;

	  synch_lookup_table[p].synch_period = temp_holder.synch_period;
	  synch_lookup_table[p].xferp_value = temp_holder.xferp_value;
	  synch_lookup_table[p].clock_dividor = temp_holder.clock_dividor;
	  synch_lookup_table[p].synch_answer = temp_holder.synch_answer;
	 }

 /* Set up initial byte counts and addresses in the tables */
 buffer_table[msg_out_buf].count = 1;
 buffer_table[msg_out_buf].address = getPhysAddr(message_out_buf);
 buffer_table[cmd_buf].count = 0;
 buffer_table[cmd_buf].address = getPhysAddr(command_buf);
 buffer_table[synch_neg_msg_out].count = 2;
 buffer_table[synch_neg_msg_out].address = getPhysAddr(sync_neg);
 buffer_table[wide_neg_msg_out].count = 1;
 buffer_table[wide_neg_msg_out].address = getPhysAddr(wide_neg);
 buffer_table[stat_buf].count = 1;
 buffer_table[stat_buf].address = getPhysAddr(check_condition);
 buffer_table[identify_msg_in_buf].count = 1;
 buffer_table[identify_msg_in_buf].address = getPhysAddr(identify_msg);
 buffer_table[msg_in_buf].count = 1;
 buffer_table[msg_in_buf].address = getPhysAddr(command_complete);
 buffer_table[save_pointers].count = 1;
 buffer_table[save_pointers].address = getPhysAddr(save_pointers_msg);
 buffer_table[disconnect_msg].count = 2;
 buffer_table[disconnect_msg].address = getPhysAddr(disc_msg);
 buffer_table[sense_data_buf].address = getPhysAddr(reset_sense_data);
 buffer_table[capacity_data_buf].count = 8;
 buffer_table[capacity_data_buf].address = getPhysAddr(read_capacity_data);

 dummy_table[msg_out_buf].count = 1;
 dummy_table[msg_out_buf].address = getPhysAddr(dummy_message_out_buf);
 dummy_table[cmd_buf].count = 0;
 dummy_table[cmd_buf].address = getPhysAddr(dummy_command_buf);
 dummy_table[stat_buf].count = 1;
 dummy_table[stat_buf].address = getPhysAddr(device_busy);
 dummy_table[msg_in_buf].count = 1;
 dummy_table[msg_in_buf].address = getPhysAddr(command_complete);
 dummy_table[sense_data_buf].address = getPhysAddr(no_sense_data);
 dummy_table[capacity_data_buf].count = 8;
 dummy_table[capacity_data_buf].address = getPhysAddr(read_capacity_data);

 /* set base scripts address and entry point for copy data */
 scripts_address = getPhysAddr(script);
 copy_data = scripts_address + Ent_copy_data;

 /* register addresses */
 dsp_address = io_port + DSP;
 dsa_address = io_port + DSA;
 dsps_address = io_port + DSPS;

 /* Set up IRQ */
 init_irq();

 /* if SCRIPTS RAM exists, copy the SCRIPTS into it */
 if ((device[selected_device].device_id == C875_DEVICE_ID)
		| ((device[selected_device].device_id == C895_DEVICE_ID)
		&& (device[selected_device].rev_id > 0x00))
		| ((device[selected_device].device_id == C825_DEVICE_ID)
		&& ((device[selected_device].rev_id & 0xf0) == C825A_VERSION_ID)))
  {
	printf("\nCopying SCRIPTS Into Onboard SCRIPTS RAM\n");

	/* get the memory base address of the SCRIPTS RAM */
	error = C8XX_GetRAMBase(&device[selected_device]);
	if (error)
	 {
	  printf("RAM Base Address Not Initialized By System\n");
	  printf("Will Continue Using SCRIPTS From System RAM\n");
	 }
	else
	 {
	  RAM_address = device[selected_device].base_addr3;

	  /* patch the SCRIPTS with the byte count and addresses */
	  script[E_count_Used[0]] &= 0xff000000L;
	  script[E_count_Used[0]] |= (sizeof(SCRIPT) & 0xffffffL);
	  script[E_source_address_Used[0]] = scripts_address;
	  script[E_destination_address_Used[0]] = RAM_address;

	  /* clear all pending interrupts */
	  do
		{
		 istat = inportb(io_port + ISTAT);
		 if (istat)
		  {
			dstat = inportb(io_port + DSTAT);
			sist0 = inportb(io_port + SIST0);
			sist1 = inportb(io_port + SIST1);
		  }
		} while (istat);

	  /* start the SCRIPT */
	  outport32(dsp_address, copy_data);

	  /* wait for the interrupt to come */
	  got_int = 0;
	  do
		{
		 enable();
		} while (!got_int);

	  /* assign the SCRIPTS address pointer to the SCRIPTS RAM base address */
	  scripts_address = RAM_address;
	  printf("SCRIPTS Will Now Be Run From Onboard SCRIPTS RAM\n");
	 }
	printf("Press Any Key To Continue\n");
	cont = getch();
  }

 /* assign the remaining entry point addresses */
 wait_select = scripts_address + Ent_wait_select;
 msg_out_phase = scripts_address + Ent_msg_out_phase;
 tur = scripts_address + Ent_tur;
 stopped_busy_tur = scripts_address + Ent_stopped_busy_tur;
 request_sense = scripts_address + Ent_request_sense;
 read_return = scripts_address + Ent_read_return;
 read_reconnect = scripts_address + Ent_read_reconnect;
 write_return = scripts_address + Ent_write_return;
 write_reconnect = scripts_address + Ent_write_reconnect;
 synch_wide_neg_return = scripts_address + Ent_synch_wide_neg_return;
 msg_in_phase = scripts_address + Ent_msg_in_phase;
 stopped_busy_wait_select = scripts_address + Ent_stopped_busy_wait_select;
 inquiry = scripts_address + Ent_inquiry;
 read_capacity = scripts_address + Ent_read_capacity;

 /* start first script */
 outport32(dsp_address, wait_select);

 /* call print_menu() */
 print_menu();

 /* as long as no menu key is pressed, this loop will simply keep enabling */
 /* system interupts.  if a key is pressed, we go into the case statement to */
 /* determine which option was chosen and handle it.  the values that had to */
 /* be used in the case statement are keyboard scan codes */
 do
  {
	enable();
	if (_bios_keybrd(_NKEYBRD_READY) != 0)
	 {
	  istat = inportb(io_port + ISTAT); /* read ISTAT */
	  istat |= 0x20; /* OR in the SIGP bit */
	  printf("\nWriting SIGP Bit To Abort SCRIPTS....\n");
	  got_int = 0;
	  outportb(io_port + ISTAT, istat); /* write ISTAT */

	  /* wait for interrupt to come back */
	  do
		{
		 enable();
		} while (!got_int);
	  ctest2 = inportb(io_port + CTEST2); /* clear the SIGP bit */
	  printf("SCRIPTS Successfully Aborted\n\n");

	  /* determine which key was pressed */
	  switch(_bios_keybrd(_NKEYBRD_READ))
		{

		 /* allows user to alter the drive cache size to force disconnects */
		 /* at different transfer sizes */
		 case 561:
		  do
			{

			 /* print out the old cache size */
			 if (data_buf_size == 0)
			  printf("Old Cache Size = 512 Bytes\n");
			 else
			  printf("Old Cache Size = %luK\n", data_buf_size / 1024);
			 printf("New Cache Size(In Kilobytes <= 1024K, 0 = 512 bytes) = ?");
			 scanf("%lu", &data_buf_size);
			}while (data_buf_size > 1024);

		  /* size of zero gives 512 byte cache, otherwise, value is in KBytes */
		  if (data_buf_size == 0)
			data_buf_size = 512;
		  else
			data_buf_size *= 1024;
		  printf("\nCache Size Modified\n");
		  printf("Press Any Key To Continue\n");
		  cont = getch();
		  break;

		 /* save ram disk from memory to a hard drive */
		 case 818:

		  /* get the file name from the user */
		  printf("Enter File Name With Path:");
		  scanf("%s", &file_name); /* get file name */

		  /* open file.  on error, print error message and break */
		  if ((save_file = fopen(file_name, "wb")) == NULL)
			{
			 printf("Error Opening File, Save Not Possible\n");
			 printf("Press Any Key To Continue\n");
			 cont = getch();
			 break;
			}
		  printf("This Will Take A Few Moments\n");

		  /* save the ram disk size and block size */
		  fprintf(save_file, "%lu %u", size, block_size);

		  /* patch SCRIPTS with the destination address and byte count */
		  script[E_destination_address_Used[0]] =
														  getPhysAddr(ptr_data_transfer_buf);
		  script[E_count_Used[0]] &= 0xff000000L;
		  script[E_count_Used[0]] |= data_transfer_buf_size;
		  addr = 0x100000L; /* initial address */
		  bytes_left = size; /* total bytes to transfer */
		  done = 0;
		  do
			{

			 /* if the number of bytes left isgreater than the data transfer */
			 /* buffer, then we will need to loop again, if not, then this is */
			 /* the last time through the loop */
			 if (bytes_left > data_transfer_buf_size)
			  {

				/* patch current source address */
				script[E_source_address_Used[0]] = addr;
				bytes_left -= data_transfer_buf_size; /* decrement count */
				addr += data_transfer_buf_size;       /* and increment */
			  }												  /* address */
			 else
			  {
				script[E_count_Used[0]] &= 0xff000000L; /* change count */
				script[E_count_Used[0]] |= bytes_left;  /* for last move */
				script[E_source_address_Used[0]] = addr;
				done = 1;
			  }
			 got_int = 0;
			 outport32(dsp_address, copy_data); /* start SCRIPT */

			 /* wait for interrupt to come back */
			 do
			  {
				enable();
			  } while (!got_int);

			 /* move data from the buffer to the hard drive */
			 if (!done)
			  fwrite(ptr_data_transfer_buf, data_transfer_buf_size, 1, save_file);
			 else
			  fwrite(ptr_data_transfer_buf,(unsigned int)bytes_left, 1, save_file);
			}while (!done);
		  fclose(save_file); /* close file */
		  printf("Ram Disk Saved\n");
		  printf("Press Any Key To Continue\n");
		  cont = getch();
		  break;

		 /* restore ram disk from a hard drive to memory */
		 case 1075:

		  /* get the file name from the user */
		  printf("Enter File Name With Path:");
		  scanf("%s", &file_name); /* get file name */

		  /* open file.  on error, print error message and break */
		  if ((save_file = fopen(file_name, "rb")) == NULL)
			{
			 printf("Error Opening File, Restore Not Possible\n");
			 printf("Press Any Key To Continue\n");
			 cont = getch();
			 break;
			}

		  /* get ram disk size and block size */
		  fscanf(save_file, "%lu %u", &file_size, &new_block_size);
		  if (file_size > size) /* if file to large, print error */
			{                    /* and break */
			 printf("File Too Large, Aborting Load\n");
			 fclose(save_file);
			 printf("Press Any Key To Continue\n");
			 cont = getch();
			 break;
			}

		  /* if the file size or block size is different than the */
		  /* current setup, the setup must be modified */
		  if ((file_size != size) || (new_block_size != block_size))
			{
			 size = file_size;
			 block_size = new_block_size;
			 last_lba = (size / block_size) - 1;
			 printf("New Drive Size = %.2f MBytes (%lu Bytes)\n",
															(((float)size) / 1048576L), size);
			 printf("New Block Size = %u\n", block_size);
			 for (i = 0, x = 3; i < 4; i++, x--)
			  read_capacity_data[i] = (int)((last_lba >> 8 * x) & 0xff);
			 for (i = 0, x = 3; i < 4; i++, x--)
			  read_capacity_data[i + 4] = (block_size >> 8 * x) & 0xff;
			}

		  printf("This Will Take A Few Moments\n");

		  /* patch SCRIPTS with the source address and byte count */
		  script[E_source_address_Used[0]] = getPhysAddr(ptr_data_transfer_buf);
		  script[E_count_Used[0]] &= 0xff000000L;
		  script[E_count_Used[0]] |= data_transfer_buf_size;
		  addr = 0x100000L; /* initial address */
		  bytes_left = size; /* total bytes to transfer */
		  done = 0;
		  do
			{

			 /* move data from the hard drive to the buffer */
			 if (bytes_left > data_transfer_buf_size)
			  fread(ptr_data_transfer_buf, data_transfer_buf_size, 1, save_file);
			 else
			  {
				fread(ptr_data_transfer_buf,(unsigned int)bytes_left, 1, save_file);
				done = 1;
			  }
			 if (!done)
			  {

				/* patch current destination address */
				script[E_destination_address_Used[0]] = addr;
				bytes_left -= data_transfer_buf_size; /* decrement count */
				addr += data_transfer_buf_size;       /* and increment */
																  /* address */
			  }
			 else
			  {
				script[E_count_Used[0]] &= 0xff000000L; /* change count */
				script[E_count_Used[0]] |= bytes_left;  /* for last move */
				script[E_destination_address_Used[0]] = addr;
			  }
			 got_int = 0;
			 outport32(dsp_address, copy_data); /* start SCRIPT */

			 /* wait for interrupt to come back */
			 do
			  {
				enable();
			  } while (!got_int);
			}while (!done);
		  fclose(save_file); /* close file */
		  printf("Ram Disk Restored\n");
		  printf("Host Should Be Reset For Proper Operation\n");
		  printf("Press Any Key To Continue\n");
		  cont = getch();
		  break;

		 /* print a logical block to the screen */
		 case 1332:

		  /* print the number of the highest accessable logical block */
		  printf("Highest Logical Block = %lx\n", last_lba);
		  do
			{
			 printf("Enter Logical Block Number (In Hex):");
			 scanf("%lx", &use_lba); /* get block number */
			}while (use_lba > last_lba); /* must be in range */

		  /* calculate source address and patch into SCRIPT */
		  addr = 0x100000L + (use_lba * block_size);
		  script[E_source_address_Used[0]] = addr;

		  /* patch buffer address and byte count */
		  script[E_destination_address_Used[0]] =
														  getPhysAddr(ptr_data_transfer_buf);
		  script[E_count_Used[0]] &= 0xff000000L;
		  script[E_count_Used[0]] |= block_size;
		  got_int = 0;
		  outport32(dsp_address, copy_data); /* start SCRIPT */

		  /* wait for interrupt to come back */
		  do
			{
			 enable();
			} while (!got_int);

		  /* clear the screen and print the logical block */
		  clrscr();
		  i = 0;
		  do
			{
			 for (p = 0; p < 16; p++) /* 16 lines per page */
			  {
				for (x = 0; x < 16; x++) /* 16 bytes per line */
				 printf("%02x ", ptr_data_transfer_buf[i + x]);/* in hex */
				printf("     ");
				for (x = 0; x < 16; x++, i++)
				 if ((ptr_data_transfer_buf[i] < 32)
				  || (ptr_data_transfer_buf[i] > 126))
				  printf(".");
				 else
				  printf("%c", ptr_data_transfer_buf[i]); /* in ascii */
				printf("\n");
			  }
			 printf("Press Any Key To Continue\n");
			 cont = getch();
			} while (i < block_size);
		  break;

		 /* modify a logical block */
		 case 1589:

		  /* print the number of the highest accessable logical block */
		  printf("Highest Logical Block = %lx\n", last_lba);
		  do
			{
			 printf("Enter Logical Block Number (In Hex):");
			 scanf("%lx", &use_lba); /* get block number */
			}while (use_lba > last_lba); /* must be in range */
		  do
			{
			 printf("Enter Start Offset:");
			 scanf("%u", &lba_offset); /* get start offset in block */
			}while (lba_offset > block_size - 1); /* must be in block */
		  do
			{
			 printf("Enter Byte Count To Modify:");
			 scanf("%u", &byte_count); /* get number of bytes to modify */

			 /* can't cross logical block boundary */
			}while (byte_count > (block_size - lba_offset));

		  inc_value = 0;
		  /* if only one byte, get new value */
		  if (byte_count == 1)
			{
			 do
			  {
				printf("Enter New Value (In Hex):");
				scanf("%x", &value);
			  }while (value > 0xff);
			}

		  /* if more than one byte, get start and increment values */
		  if (byte_count > 1)
			{
			 do
			  {
				printf("Enter Starting Value (In Hex):");
				scanf("%x", &value);
			  }while (value > 0xff);
			 printf("Enter Increment Value (In Hex):");
			 scanf("%x", &inc_value);
			}
		  for (i = 0; i < byte_count; i++) /* set up data in buffer */
			{
			 ptr_data_transfer_buf[i] = (ubyte)value;
			 value += inc_value;
			}

		  /* calculate start address and patch SCRIPT */
		  addr = 0x100000L + (use_lba * block_size) + lba_offset;
		  script[E_destination_address_Used[0]] = addr;
		  script[E_source_address_Used[0]] = getPhysAddr(ptr_data_transfer_buf);
		  script[E_count_Used[0]] &= 0xff000000L;
		  script[E_count_Used[0]] |= byte_count;
		  got_int = 0;
		  outport32(dsp_address, copy_data); /* start SCRIPT */

		  /* wait for interrupt to come back */
		  do
			{
			 enable();
			}while (!got_int);
		  printf("Logical Block Modified\n");
		  printf("Press Any Key To Continue\n");
		  cont = getch();
		  break;

		 /* read a chip register */
		 case 1846:
		  printf("Enter Register Number (In Hex):");
		  scanf("%x", &reg); /* get register number */
		  printf("\nRegister %xh = %02xh\n", reg, inportb(io_port + reg));
		  printf("Press Any Key To Continue\n");
		  cont = getch();
		  break;

		 /* modify a chip register */
		 case 2103:
		  printf("Enter Register Number (In Hex):");
		  scanf("%x", &reg); /* get register number */

		  /* print out the old register value */
		  printf("Old Value = %02xh\n", inportb(io_port + reg));
		  do
			{
			 printf("Enter New Value (In Hex):");
			 scanf("%x", &value); /* get the new value */
			}while (value > 0xff);
		  outportb(io_port + reg, value);
		  printf("\nRegister Modified\n");
		  printf("Press Any Key To Continue\n");
		  cont = getch();
		  break;

		 /* read a PCI config register */
		 case 2360:
		  do
			{
			 printf("Enter Config Register Number (In Hex):");
			 scanf("%x", &reg); /* get register number */
			}while (reg > 0x40); /* must be in range */
		  dword_offset = reg / 4; /* convert the register offset to a */
		  dword_offset *= 4;      /* DWORD aligned offset */

		  /* call the get config register function */
		  config_value = PCI_GetConfigRegister(&device[selected_device],
																					  dword_offset);

		  /* mask off the other registers in the DWORD */
		  config_value = (config_value >> (8 * (reg & 0x03))) & 0xff;
		  printf("\nConfig Register %xh = %02xh\n", reg, config_value);
		  printf("Press Any Key To Continue\n");
		  cont = getch();
		  break;

		 /* modify a PCI config register */
		 case 2617:
		  do
			{
			 printf("Enter Config Register Number (In Hex):");
			 scanf("%x", &reg); /* get register number */
			}while (reg > 0x40); /* must be in range */
		  dword_offset = reg / 4; /* convert the register offset to a */
		  dword_offset *= 4;      /* DWORD aligned offset */

		  /* get the old config register value and print it out */
		  config_value = PCI_GetConfigRegister(&device[selected_device],
																					  dword_offset);
		  printf("Old Value = %02xh\n", (config_value >>
																	(8 * (reg & 0x03))) & 0xff);

		  /* the mask is to zero out only the lane of the DWORD that */
		  /* the user wants to modify */
		  mask_value = 0;
		  for (i = 0; i < 4; i++)
			if (i != (reg & 0x03))
			 mask_value += ((unsigned long int)0xff << (8 * i));
		  config_value &= mask_value;
		  do
			{
			 printf("Enter New Value (In Hex):");
			 scanf("%x", &value); /* get the new value */
			}while (value > 0xff); /* must be only one byte wide */

		  /* the mask is now used to OR the value into the register */
		  mask_value = (unsigned long int)value << (8 * (reg & 0x03));
		  config_value |= mask_value;

		  /* call the set config register function */
		  PCI_SetConfigRegister(&device[selected_device],
																	dword_offset, config_value);
		  printf("\nRegister Modified\n");
		  printf("Press Any Key To Continue\n");
		  cont = getch();
		  break;

		 /* change SCSI ID */
		 case 5993:

		  /* print out the old SCSI ID */
		  printf("Old SCSI ID = %d\n", id);
		  do
			{
			 printf("New SCSI ID = ?"); /* get SCSI id from user */
			 scanf("%d", &id);
			}while (id > max_id);

		  /* put new id into scid register */
		  scid = (ubyte)(0x20 | (id & 0x0f));
		  outportb(io_port + SCID, scid);

		  /* the left shift puts a one in the bit that corresponds to */
		  /* the user input id */
		  respid = 1;
		  respid = (respid << id);

		  /* set response id */
		  if (wide_chip)
			outport(io_port + RESPID0, respid);
		  else
			outportb(io_port + RESPID0, (ubyte)respid);

		  printf("\nSCSI ID Changed\n");
		  printf("Host Should Be Reset For Proper Operation\n");
		  printf("Press Any Key To Continue\n");
		  cont = getch();
		  break;

		 /* quit */
		 case 4209:
		  release_irq(); /* release interrupt and restore mask */
		  exit(0); /* terminate program */
		  break;
		}

	  /* restart SCRIPTS, clear the screen, and reprint the user menu */
	  outport32(dsp_address, wait_select);

	  /* call print_menu() */
	  print_menu();
	 }
  } while (1); /* loop until the quit option is invoked */
}

/* interrupt service routine invoked by the chip interrupt pin */
static void
__interrupt isr()
{
 disable(); /* disable interrupts */
 istat = inportb(io_port + ISTAT); /* read ISTAT to determine interrupt type */
 if (istat & 1) /* this is a DMA interrupt */
  {
	dstat = inportb(io_port + DSTAT) & 0x7f; /* read DSTAT to determine type */
	switch (dstat)
	 {
	  case 0x04: /* SCRIPTS interrupt instruction */
		vector = inport32(dsps_address); /* read interrupt vector from dsps */
		switch (vector)
		 {
		  case 0x00: /* access medium for read */
			if (first == 1) /* first interrupt of the read command */
			 {
			  if (command_buf[0] == 0x08) /* this is a 6-byte read command */
				{

				 /* translate blocks to bytes */
				 bytes_left = ((unsigned long int)command_buf[4]) * block_size;

				 /* calculate the logical block address to be accessed */
				 lba = ((((unsigned long int)command_buf[1] & 0x1f) << 16)
						 | ((unsigned long int)command_buf[2] << 8)
						 | command_buf[3]);
				}
			  else /* this is a 10-byte read command */
				{

				 /* translate blocks to bytes */
				 bytes_left = (((unsigned long int)command_buf[7] << 8)
									| command_buf[8]) * block_size;

				 /* calculate the logical block address to be accessed */
				 lba = (((unsigned long int)command_buf[2] << 24)
						| ((unsigned long int)command_buf[3] << 16)
						| ((unsigned long int)command_buf[4] << 8)
						| command_buf[5]);
				}
			  first = 0;
			 }

			/* only one buffer full needed */
			if (bytes_left <= data_buf_size)
			 byte_count = bytes_left;

			/* more than one buffer full needed */
			else
			 byte_count = data_buf_size;

			/* put byte count in table */
			buffer_table[data_buf].count = byte_count;

			/* calculate address and put it in the table */
			buffer_table[data_buf].address = 0x100000L + (lba * block_size);
			bytes_left -= byte_count; /* decrement total byte count */
			if (bytes_left == 0) /* we are done with this command */
			 {
			  first = 1;
			  outportb(io_port + SCRATCHB1, 0x00); /* clear 'not done' in */
			 }                                     /* finished flag */
			else
			 {
			  lba += data_buf_size / block_size; /* increment start lba */
			  outportb(io_port + SCRATCHB1, 0x01); /* set 'not done' in */
																/* finished flag */
			 }

			/* if semaphore bit set, we are disconnected */
			if (istat & 0x10)
			 outport32(dsp_address, read_reconnect); /* start SCRIPT */
			else
			 outport32(dsp_address, read_return); /* start SCRIPT */
			break;

		  case 0x01: /* access medium for write */
			if (first == 1) /* first time for this write command */
			 {
			  if (command_buf[0] == 0x0a) /* this is a 6-byte write command */
				{

				 /* translate blocks to bytes */
				 bytes_left = ((unsigned long int)command_buf[4]) * block_size;

				 /* calculate the logical block address to be accessed */
				 lba = ((((unsigned long int)command_buf[1] & 0x1f) << 16)
						 | ((unsigned long int)command_buf[2] << 8) | command_buf[3]);
				}
			  else /* this is a 10-byte write command */
				{

				 /* translate blocks to bytes */
				 bytes_left = (((unsigned long int)command_buf[7] << 8)
									| command_buf[8]) * block_size;

				 /* calculate the logical block address to be accessed */
				 lba = (((unsigned long int)command_buf[2] << 24)
						| ((unsigned long int)command_buf[3] << 16)
						| ((unsigned long int)command_buf[4] << 8) | command_buf[5]);
				}
			  first = 0;
			 }

			/* only one buffer full needed */
			if (bytes_left <= data_buf_size)
			 byte_count = bytes_left;

			/* more than one buffer full needed */
			else
			 byte_count = data_buf_size;

			/* put byte count in table */
			buffer_table[data_buf].count = byte_count;

			/* calculate address and put it in the table */
			buffer_table[data_buf].address = 0x100000L + (lba * block_size);
			bytes_left -= byte_count; /* decrement total byte count */
			if (bytes_left == 0) /* we are done with this command */
			 {
			  first = 1;
			  outportb(io_port + SCRATCHB1, 0x10); /* set 'last time' */
																/* in finished flag */
			 }
			else
			 {
			  lba += data_buf_size / block_size; /* increment start lba */
			  outportb(io_port + SCRATCHB1, 0x11); /* set 'not done' */
																/* in finished flag */
			 }

			/* if semaphore bit set, we are disconnected */
			if (istat & 0x10)
			 outport32(dsp_address, write_reconnect); /* start SCRIPT */
			else
			 outport32(dsp_address, write_return); /* start SCRIPT */
			break;

		  case 0x02: /* disconnect on last transfer of write */
			outportb(io_port + SCRATCHB1, 0x00); /* clear 'not done' */
															 /* in finished flag */
			outport32(dsp_address, write_reconnect); /* go back and reselect */
			break;

		  case 0x03: /* this is a seek command */
			outport32(dsp_address, tur); /* go right to status and message */
			break;

		  case 0x04: /* synchronous negotiation occured */
			period = ((int)sync_neg[0]); /* get period sent by the initiator */
			offset = ((int)sync_neg[1]); /* get offset sent by the initiator */
			if (period == 0) /* they want asynchronous */
			 synch_msg_in_buf[3] = 0; /* send back a 0 */
			else

			 /* these are the new synch values that don't follow the convention */
			 /* for this section to work, they must be translated into their */
			 /* equivelent quarter periods */
			 if (period == 0x0a)
			  period = 0x06;
			 else
			  if (period == 0x0b)
				period = 0x07;

			/* go through the lookup table and find the closest value that */
			/* is equal to or greater than the request */
			p = -1;
			i = 0;
			done = 0;
			do
			 {
			  if (synch_lookup_table[i].synch_period >= (period * 4))
				{
				 p = i;
				 done = 1;
				}
			  i++;
			 }while (!done);

			if (p == -1) /* we can't go that slow */
			 offset = 0; /* go asynchronously */

			if (offset > max_offset) /* initiator request is too big, send back */
			 offset = max_offset;    /* our best offset */
			synch_msg_in_buf[4] = ((ubyte)offset); /* set offset in buffer */

			if (p != -1)
			 {
			  /* set up the synch neg answer */
			  synch_msg_in_buf[3] = synch_lookup_table[p].synch_answer;

			  /* switch cannot use float, so convert to int */
			  x = (int)synch_lookup_table[p].clock_dividor;
			  if (synch_lookup_table[p].clock_dividor == 1.5)
				x = 5;

			  /* zero out the synchronous clock divider in SCTNL3 */
			  scntl3 = inportb(io_port + SCNTL3) & 0x0f;

			  /* if synch period is less than 75ns, set the fast-20/fast-40 bit */
			  if (synch_lookup_table[p].synch_period < 75)
				scntl3 |= 0x80;

			  switch (x) /* set up the synchronous clock divider */
				{
				 case 1: scntl3 |= 0x10;
							break;
				 case 2: scntl3 |= 0x30;
							break;
				 case 3: scntl3 |= 0x40;
							break;
				 case 4: scntl3 |= 0x50;
							break;
				 case 5: scntl3 |= 0x20;
							break;
				 case 6: scntl3 |= 0x60;
							break;
				 case 8: scntl3 |= 0x70;
							break;
				}
			  outportb(io_port + SCNTL3, scntl3);

			  /* set up sxfer */
			  sxfer = ((synch_lookup_table[p].xferp_value << 5) | offset);

			  /* after synchronous negotiation, we must update the table info */
			  sel_info[3] = scntl3;
			  sel_info[1] = sxfer;
			  buffer_table[selector_id].count =
													 (((unsigned long int)sel_info[3] << 24)
													| ((unsigned long int)sel_info[2] << 16)
													| ((unsigned long int)sel_info[1] << 8)
													| sel_info[0]);

			  outportb(io_port + SXFER, sxfer);

			  /* if we haven't done wide negotiation yet, clear the screen and */
			  /* reprint the user menu before giving the synch results */
			  if ((!done_wide) || (done_synch))
				{

				 /* call print_menu() */
				 print_menu();

				 if (done_synch)
				  done_wide = 0;
				}
			  printf("\nSynchronous Negotiation Occured\n");

			  if (offset == 0)
				printf("Running Asynchronously\n");
			  else

				/* print out the theoretical maximum transfer rate */
				printf("Max Transfer Rate = %.2f MBytes/Sec\n",
							((1 / (float)synch_lookup_table[p].synch_period) * 1000) *
							((float)((scntl3 & 0x08) >> 3) + 1));
			  done_synch = 1;
			 }

			/* set up the byte count and address of the synch neg answer in the */
			/* table indirect table */
			buffer_table[neg_msg_in].count = 5;
			buffer_table[neg_msg_in].address = getPhysAddr(synch_msg_in_buf);
			outport32(dsp_address, synch_wide_neg_return); /* start SCRIPT */
			break;

		  case 0x05: /* wide negotiation occured */

			/* if we haven't done synch negotiation yet, clear the screen and */
			/* reprint the user menu before giving the wide negotiation results */
			if ((!done_synch) || (done_wide))
			 {

			  /* call print_menu() */
			  print_menu();

			  if (done_wide)
				done_synch = 0;
			 }

			/* narrow request */
			if ((wide_neg[0] == 0x00) | (wide_chip == 0))
			 {
			  printf("\nWide Negotiation Occured\n");
			  printf("Running Narrow\n");
			  wide_msg_in_buf[3] = 0x00; /* set narrow in buffer */

			  /* set up the byte count and address of the wide neg answer in */
			  /* the table indirect table */
			  buffer_table[neg_msg_in].count = 4;
			  buffer_table[neg_msg_in].address = getPhysAddr(wide_msg_in_buf);
			  scntl3 = inportb(io_port + SCNTL3); /* read the scntl3 register */
			  scntl3 &= 0xf7; /* clear wide SCSI bit */

			  /* after wide negotiation, we must update the table info */
			  sel_info[3] = scntl3;
			  buffer_table[selector_id].count =
													 (((unsigned long int)sel_info[3] << 24)
													| ((unsigned long int)sel_info[2] << 16)
													| ((unsigned long int)sel_info[1] << 8)
													| sel_info[0]);

			  /* if synchronous negotiation has already occured, update the */
			  /* theoretical maximum transfer rate */
			  if (done_synch && (offset != 0))
				printf("\nMax Transfer Rate = %.2f MBytes/Sec\n",
							((1 / (float)synch_lookup_table[p].synch_period) * 1000) *
							((float)((scntl3 & 0x08) >> 3) + 1));
			  outportb(io_port + SCNTL3, scntl3); /* write value to register */
			  outport32(dsp_address, synch_wide_neg_return); /* start SCRIPT */
			  done_wide = 1;
			 }
			else

			 /* wide or double wide request */
			 if ((wide_neg[0] == 0x01) | (wide_neg[0] == 0x02))
			  {
				printf("\nWide Negotiation Occured\n");
				printf("Running Wide\n");
				wide_msg_in_buf[3] = 0x01; /* set wide in buffer */

				/* set up the byte count and address of the wide neg answer in */
				/* the table indirect table */
				buffer_table[neg_msg_in].count = 4;
				buffer_table[neg_msg_in].address = getPhysAddr(wide_msg_in_buf);
				scntl3 = inportb(io_port + SCNTL3); /* read scntl3 register */
				scntl3 |= 0x08; /* set wide SCSI bit */

				/* after wide negotiation, we must update the table info */
				sel_info[3] = scntl3;
				buffer_table[selector_id].count =
													 (((unsigned long int)sel_info[3] << 24)
													| ((unsigned long int)sel_info[2] << 16)
													| ((unsigned long int)sel_info[1] << 8)
													| sel_info[0]);

				/* if synchronous negotiation has already occured, update the */
				/* theoretical maximum transfer rate */
				if (done_synch && (offset != 0))
				 printf("\nMax Transfer Rate = %.2f MBytes/Sec\n",
							((1 / (float)synch_lookup_table[p].synch_period) * 1000) *
							((float)((scntl3 & 0x08) >> 3) + 1));
				outportb(io_port + SCNTL3, scntl3); /* write value to register */
				outport32(dsp_address, synch_wide_neg_return); /* start SCRIPT */
				done_wide = 1;
			  }
			 else /* invalid wide negotiation message */
			  {

				/* put address of message reject message in table */
				buffer_table[msg_in_buf].address = getPhysAddr(message_reject);
				outport32(dsp_address, msg_in_phase); /* start SCRIPT */
				return_code = 1; /* says this was rejecting a bad message */
			  }
			break;

		  case 0x06: /* a message not handled by SCRIPTS was received */
			switch(message_out_buf[0])
			 {
			  case 0x0c: /* bus device reset message */

				/* call print_menu */
				print_menu();

				printf("\nBus Device Reset Message Received\n");

				/* call device_reset */
				device_reset();

				outport32(dsp_address, wait_select); /* start SCRIPT */
				break;

			  case 0x05: /* initiator detected error message */

				/* send back a restore pointers message to retry the transaction */
				buffer_table[msg_in_buf].address =
															getPhysAddr(restore_pointers_msg);

				/* after sending the message, resume the last operation */
				return_code = 2;
				outport32(dsp_address, msg_in_phase); /* start SCRIPT */
				break;

			  case 0x07: /* we got a message reject -should never happen- */
				release_irq(); /* release interrupt and restore mask */
				exit(1); /* terminate program with error code */
				break;

			  case 0x09: /* we got a message parity error message */
				buffer_table[msg_in_buf].address =
															getPhysAddr(restore_pointers_msg);

				/* after sending the message, resume the last operation */
				return_code = 2;
				outport32(dsp_address, msg_in_phase); /* start SCRIPT */
				break;

			  default: /* any other message gets a message reject */

				/* put address of message reject message in table */
				buffer_table[msg_in_buf].address = getPhysAddr(message_reject);
				outport32(dsp_address, msg_in_phase); /* start SCRIPT */

				/* after sending the message, return to message out phase */
				return_code = 1;
				break;
			 }
			break;

		  case 0x07: /* a bad extended message was received */

			/* put address of message reject message in table */
			buffer_table[msg_in_buf].address = getPhysAddr(message_reject);
			outport32(dsp_address, msg_in_phase); /* start SCRIPT */

			/* after sending the message, return to message out phase */
			return_code = 1;
			break;

		  case 0x08: /* we sent a message to the initiator */

			/* the return code is based on what happened that caused the */
			/* message to be sent */
			switch (return_code)
			 {
			  case 1: /* this was rejecting a bad message */

				/* restore command complete to the message in buffer */
				buffer_table[msg_in_buf].address = getPhysAddr(command_complete);
				outport32(dsp_address, msg_out_phase); /* start SCRIPT */
				break;

			  case 2: /* this was from a parity error */

				/* restore command complete to the message in buffer */
				buffer_table[msg_in_buf].address = getPhysAddr(command_complete);

				/* start SCRIPTS at the location where the error was detected */
				outport32(dsp_address, dsp);
				break;
			 }
			break;

		  case 0x09: /* got a request sense command */

			/* if the code is busy working on another job, the regular table */
			/* can be used, otherwise, the dummy table must be used so as not */
			/* to mess up the table entries that pertain to the previous job */
			if (!busy)
			 {

			  /* if the initiator asked for less sense data than we have, send */
			  /* back only the amount asked for */
			  if (command_buf[4] <= 0x12)
				buffer_table[sense_data_buf].count = command_buf[4];

			  /* if the initiator asked for more sense data than we have, send */
			  /* back only what we have */
			  else
				buffer_table[sense_data_buf].count = 0x12;

			  /* if stopped, we must still send back good status */
			  if (return_flag == 4)
				buffer_table[stat_buf].address = getPhysAddr(status_good);
			 }
			else
			 {

			  /* if the initiator asked for less sense data than we have, send */
			  /* back only the amount asked for */
			  if (dummy_command_buf[4] <= 0x12)
				dummy_table[sense_data_buf].count = dummy_command_buf[4];

			  /* if the initiator asked for more sense data than we have, send */
			  /* back only what we have */
			  else
				dummy_table[sense_data_buf].count = 0x12;

			  /* status must be good, even if busy */
			  dummy_table[stat_buf].address = getPhysAddr(status_good);
			 }
			outport32(dsp_address, request_sense); /* start SCRIPT */
			break;

		  case 0x0a: /* got an inquiry command */

			/* if the code is busy working on another job, the regular table */
			/* can be used, otherwise, the dummy table must be used so as not */
			/* to mess up the table entries that pertain to the previous job */
			if (!busy)
			 {

			  /* find out what the initiator's id is and put that in the table */
			  /* this assumes that an inquiry will be done before the initiator */
			  /* attempts to transfer data with disconnects and that only one */
			  /* initiator is operating on the SCSI bus */
			  ssid = inportb(io_port + SSID) & 0x0f;
			  sel_info[2] = ssid;
			  buffer_table[selector_id].count =
													 (((unsigned long int)sel_info[3] << 24)
													| ((unsigned long int)sel_info[2] << 16)
													| ((unsigned long int)sel_info[1] << 8)
													| sel_info[0]);

			  /* if the initiator asked for less inquiry data than we have, send */
			  /* back only the amount asked for */
			  if (command_buf[4] <= 0x24)
				buffer_table[inquiry_data_buf].count = command_buf[4];

			  /* if the initiator asked for more inquiry data than we have, send */
			  /* back only what we have */
			  else
				buffer_table[inquiry_data_buf].count = 0x24;

			  /* if this command is for a valid lun, send back the normal data */
			  if ((inportb(io_port + SCRATCHB0) & 0x07) == 0)
				buffer_table[inquiry_data_buf].address = getPhysAddr(inquiry_data);

			  /* if this command is for an invalid lun, send back bad lun data */
			  else
				buffer_table[inquiry_data_buf].address =
														  getPhysAddr(bad_lun_inquiry_data);

			  /* if a check condition exists, we must still send good status */
			  if (return_flag == 2)
				{
				 buffer_table[stat_buf].address = getPhysAddr(status_good);

				 /* after command complete, return status to check condition */
				 return_flag=3;
				}

			  /* if stopped, we must still send back good status */
			  if (return_flag == 4)
				buffer_table[stat_buf].address = getPhysAddr(status_good);
			 }
			else
			 {

			  /* if the initiator asked for less inquiry data than we have, send */
			  /* back only the amount asked for */
			  if (dummy_command_buf[4] <= 0x24)
				dummy_table[inquiry_data_buf].count = dummy_command_buf[4];

			  /* if the initiator asked for more inquiry data than we have, send */
			  /* back only what we have */
			  else
				dummy_table[inquiry_data_buf].count = 0x24;

			  /* if this command is for a valid lun, send back the normal data */
			  if ((message_out_buf[0] & 0x07) == 0)
				dummy_table[inquiry_data_buf].address = getPhysAddr(inquiry_data);

			  /* if this command is for an invalid lun, send back bad lun data */
			  else
				dummy_table[inquiry_data_buf].address =
														  getPhysAddr(bad_lun_inquiry_data);

			  /* status must be good, even if busy */
			  dummy_table[stat_buf].address = getPhysAddr(status_good);
			 }
			outport32(dsp_address, inquiry); /* start SCRIPT */
			break;

		  case 0x0b: /* read capacity command */

			/* if the code is busy working on another job, the regular table */
			/* can be used, otherwise, the dummy table must be used so as not */
			/* to mess up the table entries that pertain to the previous job */
			if (!busy)
			 {

			  /* if stopped, we must still send back good status */
			  if (return_flag == 4)
				buffer_table[stat_buf].address = getPhysAddr(status_good);
			 }
			else

			  /* status must be good, even if busy */
			  dummy_table[stat_buf].address = getPhysAddr(status_good);
			outport32(dsp_address, read_capacity); /* start SCRIPT */
			break;

		  case 0x0c: /* got a start/stop command */
			if (command_buf[4] & 1) /* This is a start command */
			 {

			  /* restore good status to buffer */
			  buffer_table[stat_buf].address = getPhysAddr(status_good);

			  /* clear the sense data buffer */
			  buffer_table[sense_data_buf].address=getPhysAddr(no_sense_data);
			  return_flag = 0; /* normal end of command */
			  outportb(io_port + SCRATCHA0, 0x00); /* clear stopped flag */
			  outportb(io_port + SCRATCHA2, 0x00); /* clear stopped/busy flag */
			  outport32(dsp_address, tur); /* start SCRIPT */
			 }
			else /* this is a stop command */
			 {
			  return_flag = 1; /* at command complete, status to not ready */
			  outportb(io_port + SCRATCHA0, 0x01); /* set stopped flag */
			  outportb(io_port + SCRATCHA2, 0x01); /* set stopped/busy flag */
			  outport32(dsp_address, stopped_busy_tur); /* start SCRIPT */
			 }
			break;

		  case 0x0d: /* this is a format unit command */

			/* clear out the data transfer buffer */
			for (m = 0; m < data_transfer_buf_size; m++)
			 ptr_data_transfer_buf[m] = 0;

			/* patch SCRIPTS with source address and byte count */
			script[E_source_address_Used[0]] = getPhysAddr(ptr_data_transfer_buf);
			script[E_count_Used[0]] &= 0xff000000L;
			script[E_count_Used[0]] |= data_transfer_buf_size;
			addr = 0x100000L; /* initial address */
			bytes_left = size; /* total bytes to transfer */
			done = 0;
			outportb(io_port + DIEN, 0x00); /* disable DMA interrupts */
			outportb(io_port + SIEN0, 0x00); /* disable SCSI interrupts */
			do
			 {
			  if (bytes_left > data_transfer_buf_size)
				{

				 /* patch current destination address */
				 script[E_destination_address_Used[0]] = addr;
				 bytes_left -= data_transfer_buf_size; /* decrement count */
				 addr += data_transfer_buf_size;       /* and increment address */
				}
			  else
				{
				 script[E_count_Used[0]] &= 0xff000000L; /* change count and */
				 script[E_count_Used[0]] |= bytes_left; /* address for final move */
				 script[E_destination_address_Used[0]] = addr;
				 done = 1;
				}
			  outport32(dsp_address, copy_data); /* start SCRIPT */
			  do
				{
				 istat = inportb(io_port + ISTAT);
				} while (!(istat & 1)); /* wait for DMA interrupt */
			  dstat = inportb(io_port + DSTAT); /* get type of DMA interrupt */
			  if (!(dstat & 0x04)) /* this would be very bad */
				{
				 release_irq(); /* release interrupt and restore mask */
				 exit(1);
				}
			  vector = inport32(dsps_address); /* get interrupt vector */
			  if (vector != 0x11) /* this would also be very bad */
				{
				 release_irq(); /* release interrupt and restore mask */
				 exit(1);
				}
			 }while (!done); /* loop until done */
			outportb(io_port + DIEN, 0x75); /* enable DMA interrupts */
			outportb(io_port + SIEN0, 0x8f); /* enable SCSI interrupts */

			/* if a check condition exists, we must still send back good status */
			if (return_flag == 2)
			 {
			  buffer_table[stat_buf].address = getPhysAddr(status_good);

			  /* after command complete, return status to check condition */
			  return_flag=3;
			 }
			outport32(dsp_address, tur); /* start SCRIPT */
			break;

		  case 0x0e: /* this is a send diagnostic command */
			if (command_buf[1] & 0x04) /* self test, return good status */
			 outport32(dsp_address, tur);
			else /* this is an unsupported version of this command */
			 {

			  /* put check condition in status buffer */
			  buffer_table[stat_buf].address = getPhysAddr(check_condition);
			  return_flag = 2; /* at command complete, status back to good */

			  /* put illegal command sense data in the sense data buffer */
			  buffer_table[sense_data_buf].address = getPhysAddr(illegal_command);
			  outport32(dsp_address, tur); /* start SCRIPT */
			 }
			break;

		  case 0x0f: /* received an abort message and disconnected */
			busy = 0; /* clear the driver busy flag */
			outportb(io_port + SCRATCHB1, 0x00); /* clear the not done flag */
			outportb(io_port + SCRATCHA1, 0x00); /* clear the SCRIPTS busy flag */
			outport32(dsp_address, wait_select); /* start SCRIPT */
			if (return_flag != 4)
			 return_flag = 0; /* clear the return flag */
			break;

		  case 0x10: /* this is an illegal command */

			/* put check condition in status buffer */
			buffer_table[stat_buf].address = getPhysAddr(check_condition);
			return_flag = 2; /* at command complete, status back to good */

			/* put illegal command sense data in the sense data buffer */
			buffer_table[sense_data_buf].address = getPhysAddr(illegal_command);
			outport32(dsp_address, tur); /* start SCRIPT */
			break;

		  case 0x11: /* got SIGP and stopped */

			/* set the flag that signals receipt of interrupt */
			got_int = 1;
			break;

		  case 0x12: /* done with memory move */

			/* set the flag that signals receipt of interrupt */
			got_int = 1;
			break;

		  case 0x13: /* got selected during reselection */
			dsa = getPhysAddr(dummy_table); /* get the address of dummy table */
			outport32(dsa_address, dsa); /* put the address in the dsa register */
			busy = 1; /* set the driver busy flag */
			outportb(io_port + SCRATCHA1, 0x01); /* set the SCRIPTS busy flag */
			outportb(io_port + SCRATCHA2, 0x01); /* set stopped/busy flag */
			outport32(dsp_address, stopped_busy_wait_select); /* start SCRIPT */
			break;

		  case 0x14: /* done with command while busy */

			/* restore device busy to the status buffer in case it changed */
			dummy_table[stat_buf].address = getPhysAddr(device_busy);
			dsa = getPhysAddr(buffer_table); /* get the address of the table */
			outport32(dsa_address, dsa); /* restore table address to the dsa */
			busy = 0; /* clear the driver busy flag */
			outportb(io_port + SCRATCHA1, 0x00); /* clear the SCRIPTS busy flag */
			outportb(io_port + SCRATCHA2, 0x00); /* clear stopped/busy flag */

			/* the first byte of the command will tell where to return */
			switch(command_buf[0])
			 {
			  case 0x08: /* we were performing a 6-byte read */
			  case 0x28: /* we were performing a 10-byte read */
				outport32(dsp_address, read_reconnect); /* start SCRIPT */
				break;

			  case 0x0a: /* we were performing a 6-byte write */
			  case 0x2a: /* we were performing a 10-byte write */
				outport32(dsp_address, write_reconnect); /* start SCRIPT */
				break;
			 }
		 }
		break;

	  case 0x01: /* illegal instruction interrupt */
		printf("An Illegal Instruction Was Detected At SCRIPTS Offset ");

		/* print address of instruction that caused fault */
		addr = inport32(io_port + DSP);
		addr = addr - scripts_address;
		printf("%lxh\n", addr);
		printf("Note:  The DSP Points At The Next Dword To Be Fetched\n");

		/* print out the first dword of the instruction that caused the fault */
		dcmd_dbc = inport32(io_port + DBC);
		dsps = inport32(io_port + DSPS);
		dcmd = (ubyte)((dcmd_dbc >> 24) & 0xff);
		dbc = dcmd_dbc & 0xffffffL;
		printf("DCMD = %02xh\nDBC = %06lxh\nDSPS = %08lxh\n", dcmd, dbc, dsps);
		if (dcmd == 0xc0)
		 {
		  printf("Source Address = %08lxh\n", script[E_source_address_Used[0]]);
		  printf("Destination Address = %08lxh\n",
													 script[E_destination_address_Used[0]]);
		 }
		outportb(PIC2_EOI_REG, 0x20); /* issue the end of interrupt commands */
		outportb(PIC1_EOI_REG, 0x20);
		enable(); /* re-enable interrupts */
		release_irq(); /* release interrupt and restore mask */
		exit(1); /* terminate program with error code */
		break;

	  case 0x20: /* bus fault interrupt */
		printf("A Bus Fault Occured\n");
		outportb(PIC2_EOI_REG, 0x20); /* issue the end of interrupt commands */
		outportb(PIC1_EOI_REG, 0x20);
		enable(); /* re-enable interrupts */
		release_irq(); /* release interrupt and restore mask */
		exit(1); /* terminate program with error code */
		break;
	 }
  }
 if (istat & 4) /* this is an interrupt on the fly that signifies end of i/o */
  {
	istat = 0x04; /* clear all bits except the intfly bit */
	outportb(io_port + ISTAT, istat); /* write back istat to clear the intfly */

	/* return flag determines the action to be taken at the end of the i/o */
	switch (return_flag)
	 {
	  case 0: /* normal end of command */
		break;

	  case 1: /* this is after a stop command terminates */

		/* put check condition in the status buffer */
		buffer_table[stat_buf].address = getPhysAddr(check_condition);

		/* put unit not ready sense data in the sense data buffer */
		buffer_table[sense_data_buf].address = getPhysAddr(unit_not_ready);
		return_flag = 4; /* set the return flag for future commands */
		break;

	  case 2: /* this is after sending a check condition */

		/* after sending check condition, status goes back to good */
		buffer_table[stat_buf].address = getPhysAddr(status_good);
		return_flag = 0; /* clear the return flag */
		break;

	  case 3: /* this is after inquiry or format when check condition existed */

		/* put the check condition back in the status buffer */
		buffer_table[stat_buf].address = getPhysAddr(check_condition);
		return_flag = 2; /* put the return flag back to the way it was */
		break;

	  case 4: /* any command while stopped */

		/* if a command that is legal while stopped is received, the status */
		/* sent back was good, and now we have to put the check condition */
		/* back into the status buffer and the sense data must return to what */
		/* it was before if it changed */
		buffer_table[stat_buf].address = getPhysAddr(check_condition);
		buffer_table[sense_data_buf].address = getPhysAddr(unit_not_ready);
		break;
	 }
  }
 if (istat & 2) /* this is a SCSI interrupt */
  {
	sist0 = inportb(io_port + SIST0); /* get what SCSI interrupt occured */
	sist1 = inportb(io_port + SIST1);
	if (sist0 & 0x80) /* ATN is active on the SCSI bus */
	 {

	  /* save the instruction address of the command that caused the fault */
	  dsp = (inport32(io_port + DSP)) - 8;

	  /* check to see if there is data in the FIFO, if there is, clear it */
	  dstat = inportb(io_port + DSTAT);
	  if (!(dstat & 0x80))
		{
		 ctest3 = inportb(io_port + CTEST3);
		 ctest3 |= 0x04;
		 outportb(io_port + CTEST3, ctest3);
		 ctest3 &= 0x0b;
		 outportb(io_port + CTEST3, ctest3);
		}

	  /* go to message out phase to see what the initiator wanted */
	  outport32(dsp_address, msg_out_phase);
	 }
	if (sist0 & 1) /* a SCSI parity error was detected */
	 {

	  /* determine the phase to know how best to handle the error */
	  phase = inportb(io_port + DCMD) & 0x07;

	  /* if a parity error occures during message out phase, there is no */
	  /* defined way to recover, so reject the message */
	  if (phase == 0x06)
		{

		 /* put message reject in the message in buffer */
		 buffer_table[msg_in_buf].address = getPhysAddr(message_reject);

		 /* after sending the message reject, says we should go to message out */
		 return_code = 1;
		}

	  /* in all other cases, we should be able to do a restore pointers and */
	  /* try the transfer again */
	  else
		{

		 /* put restore pointers in the message in buffer */
		 buffer_table[msg_in_buf].address = getPhysAddr(restore_pointers_msg);

		 /* after sending the restore pointers, says we should retry the last */
		 /* instruction */
		 return_code = 2;

		 /* save the instruction address of the command that was running when */
		 /* the parity error was received */
		 dsp = (inport32(dsp_address)) - 8;
		}
	  outport32(dsp_address, msg_in_phase); /* start SCRIPT */
	 }
	if (sist0 & 2) /* a SCSI bus reset has occured */
	 {

	  /* call print_menu() */
	  print_menu();

	  printf("\nSCSI Bus Reset Received\n");

	  /* call device_reset */
	  device_reset();

	  outport32(dsp_address, wait_select); /* start SCRIPT */
	 }
  }
 outportb(PIC2_EOI_REG, 0x20); /* issue the two end of interrupt commands */
 outportb(PIC1_EOI_REG, 0x20);
 enable(); /* re-enable interrupts */
}

static void
init_siop()
{
 ubyte scntl1, gpio3, stest1, stest3;

 /* write the chip reset bit into the istat register and then clear it */
 istat = inportb(io_port + ISTAT);
 istat |= 0x40;
 outportb(io_port + ISTAT, istat);
 istat ^= 0x40;
 outportb(io_port + ISTAT, istat);

 outportb(io_port + SCNTL0, 0xc9); /* set target mode, parity checking */
 outportb(io_port + DCNTL, 0x01); /* set 700 compatibility */
 scntl1 = inportb(io_port + SCNTL1);
 scntl1 |= 0x20;
 outportb(io_port + SCNTL1, scntl1); /* set DHP */
 scid = (ubyte)(0x20 | (id & 0x0f));
 outportb(io_port + SCID, scid); /* set ID, enable response */
											/* to selection */

 /* the left shift puts a one in the bit that corresponds to the user input id*/
 respid = 1;
 respid = (respid << id);
 if (wide_chip)
  outport(io_port + RESPID0, respid); /* set response ID */
 else
  outportb(io_port + RESPID0, (ubyte)respid); /* set response ID */
 outportb(io_port + SXFER, 0x00); /* set asynchronous */
 outportb(io_port + STEST3, 0x90); /* enable tolerant and disable single*/
											  /* initiator response */

 outportb(io_port + DIEN, 0x75);	/* enable DMA interrupts */
 outportb(io_port + SIEN0, 0x8f); /* enable SCSI interrupts */

 gpio3 = inportb(io_port + GPREG) & 0x08; /* get the status of GPIO 3 */

 /* if gpio3 is pulled low, then this is a differential board */
 if (!gpio3)
  outportb(io_port + STEST2, 0x20); /* enable differential operation */

 /* if the user has selected to enable the clock doubler or quadrupler, we */
 /* need to start up the clock multiplier and halt the scsi clock before */
 /* enabling use of the multiplied clock */
 if (doubler || quadrupler)
  {
	stest1 = inportb(io_port + STEST1);
	stest1 |= 0x08;
	outportb(io_port + STEST1, stest1); /* set clock multiplier enable bit */
	delay(1); /* wait 1 ms */
	stest3 = inportb(io_port + STEST3);
	stest3 |= 0x20;
	outportb(io_port + STEST3, stest3); /* halt the SCSI clock */
  }

 /* set up clock dividers based on the SCLK value provided by user */
 if ((clk > 16) && (clk <= 25))
  scntl3 = 0x11;
 else
  if ((clk > 25) && (clk <= 38))
	scntl3 = 0x22;
  else
	if ((clk > 38) && (clk <= 50))
	 scntl3 = 0x33;
	else
	 if ((clk > 50) && (clk <= 75))
	  scntl3 = 0x44;
	 else
	  if ((clk > 75) && (clk <= 80))
		scntl3 = 0x55;
	  else
		if ((clk > 80) && (clk <=120))
		 scntl3 = 0x66;
		else
		 if ((clk > 120) && (clk <= 180))
		  scntl3 = 0x77;

 outportb(io_port + SCNTL3, scntl3); /* write the value to scntl3 */

 /* if the user has selected to enable the clock doubler or quadrupler, we */
 /* must now set the multiplied clock select bit, and restart the scsi clock */
 if (doubler || quadrupler)
  {
	stest1 = inportb(io_port + STEST1);
	stest1 |= 0x04;
	outportb(io_port + STEST1, stest1); /* set the multiplier select bit */
	stest3 = inportb(io_port + STEST3);
	stest3 &= 0xdf;
	outportb(io_port + STEST3, stest3); /* re-start the SCSI clock */
  }

 /* set up initial values in the selector id entry of the table */
 sel_info[3] = scntl3;
 sel_info[1] = 0;
 buffer_table[selector_id].count = (((unsigned long int)sel_info[3] << 24)
											 | ((unsigned long int)sel_info[2] << 16)
											 | ((unsigned long int)sel_info[1] << 8)
											 | sel_info[0]);

 /* initialize the scratcha register */
 outport32(io_port + SCRATCHA, 0);

 /* the following sections determine what chip is in use and take advantage */
 /* of the available features */

 /* this is a 53c810 */
 if ((device[selected_device].device_id == C810_DEVICE_ID)
  && ((device[selected_device].rev_id & 0xf0) == C810_VERSION_ID))
  {
	outportb(io_port + DMODE, 0x88); /* set 8-transfer burst and read line */
	max_cd = 4; /* 810 has 4 clock dividor values */
  }

 /* this is a 53c815 */
 if (device[selected_device].device_id == C815_DEVICE_ID)
  {
	outportb(io_port + DMODE, 0x0a); /* set 2-transfer burst, read line, */
												/* and burst op code */
	max_cd = 4; /* 815 has 4 clock dividor values */
  }

 /* this is a 53c825 */
 if ((device[selected_device].device_id == C825_DEVICE_ID)
		&& ((device[selected_device].rev_id & 0xf0) == C825_VERSION_ID))
  {
	outportb(io_port + DMODE, 0x8a); /* set 8-transfer burst, read line, */
												/* and burst op code */
	max_cd = 4; /* 825 has 4 clock dividor values */
  }

 /* this is either a 53c810a or a 53c860 */
 if (((device[selected_device].device_id == C810_DEVICE_ID)
		 && ((device[selected_device].rev_id & 0xf0) == C810A_VERSION_ID))
		 | (device[selected_device].device_id ==C810AP_DEVICE_ID)
		 | (device[selected_device].device_id == C860_DEVICE_ID))
  {
	outportb(io_port + CTEST3, 0x01); /* set write and invalidate */
	outportb(io_port + DMODE, 0xce); /* set 16-transfer burst, read line, */
												/* read multiple, burst op code */
	outportb(io_port + DCNTL, 0xa1); /* cache line size enable, */
												/* pre-fetch enable and 700 comp */
	if(device[selected_device].device_id == C860_DEVICE_ID)
	 max_cd = 5; /* 860 has 5 clock dividor values */
	else
	 max_cd = 4; /* 810a has 4 clock dividor values */
  }

 /* this is either a 53c825a, 53c875, or a 53c895 */
 if (((device[selected_device].device_id == C825_DEVICE_ID)
		 && ((device[selected_device].rev_id & 0xf0) == C825A_VERSION_ID))
		 | (device[selected_device].device_id == C875_DEVICE_ID)
		 | (device[selected_device].device_id == C895_DEVICE_ID))
  {
	outportb(io_port + CTEST3, 0x01); /* set write and invalidate */
	outportb(io_port + CTEST5, 0x24); /* set DMA FIFO to 536 and */
	outportb(io_port + DMODE, 0x4e); /* 64-transfer burst, read line, */
												/* read multiple, burst op code */
	outportb(io_port + DCNTL, 0x81); /* cache line size enable and 700 comp*/
	max_offset = 16; /* max synch offset in 825a and 875 is 16 */
	if (device[selected_device].device_id == C895_DEVICE_ID)
	 {
	  max_cd = 7; /* 895 has 7 clock dividor values */
	  max_offset = 31; /* max synch offset in 895 is 31 */
	 }
	else
	 if (device[selected_device].device_id == C875_DEVICE_ID)
	  max_cd = 5; /* 875 has 5 clock dividor values */
	 else
	  max_cd = 4; /* 825a has 4 clock dividor values */
  }
}

static void
init_table()
{
 /* Allocate space for the two tables of addresses used by the table indirect */
 /* commands.  These tables must be on 32-bit boundaries. */
 uquad dsa;
 uword dsa_address;

 buffer_table = (table *) my_malloc((10+TABLE_SIZE) * sizeof(table));
 if (FP_OFF(buffer_table) & 0x03)
  {
	printf("Buffer Table Not DWord Aligned, Fix Frogram\n");
	exit(1);
  }

 dummy_table = (table *) my_malloc((10+TABLE_SIZE) * sizeof(table));
 if (FP_OFF(dummy_table) & 0x03)
  {
	printf("Dummy Table Not DWord Aligned, Fix Frogram\n");
	exit(1);
  }

 dsa_address = io_port + DSA; /* set up the dsa register address pointer */
 dsa = getPhysAddr(buffer_table); /* get the physical address of the table */
 outport32(dsa_address, dsa); /* put the table address into the dsa register */
}

static void
init_script()
{
 /* Make sure SCRIPTS start on a 32-bit boundary */
 if (FP_OFF(SCRIPT) & 0x03)
  {

	/* malloc a new buffer to put the scripts in and check it's alignment */
	script = (uquad *) my_malloc(sizeof(SCRIPT));
	printf("Aligning SCRIPTS\n");
	if (FP_OFF(script) & 0x03)
	 {
	  printf("Not Able To Align SCRIPTS, Fix Program\n");
	  exit(1);
	 }

	/* move the SCRIPTS into the new buffer that is dword aligned */
	memcpy(script, SCRIPT, sizeof(SCRIPT));
  }
 else
  script = SCRIPT; /* assign the pointer to the origional buffer */
}

static void
print_menu()
{

 /* clear the screen and print the user menu */
 clrscr();
 printf("Main Menu:\n");
 printf("1. Alter Disk Cache Size\n");
 printf("2. Save Ram Disk To Hard Drive\n");
 printf("3. Restore Ram Disk From Hard Drive\n");
 printf("4. Print A Logical Block To The Screen\n");
 printf("5. Modify A Logical Block (Not Recommended In Normal Operation)\n");
 printf("6. Read A Chip Register\n");
 printf("7. Modify A Chip Register (Not Recommended In Normal Operation)\n");
 printf("8. Read a PCI Configuration Register\n");
 printf("9. Modify a PCI Configuration Register (Not Recommended In Normal ");
 printf("Operation)\n");
 printf("i. Change SCSI ID (Not Recommended In Normal Operation)\n");
 printf("q. Quit\n");
 printf("Selecting Any Of These Options Will Halt SCRIPTS\n");
}

static void
device_reset()
{

 /* call the function to reset the chip and set up initial values */
 init_siop();

 /* get the table address and put it in the DSA register */
 dsa = getPhysAddr(buffer_table);
 outport32(dsa_address, dsa);

 /* put the check condition in the status buffer */
 buffer_table[stat_buf].address = getPhysAddr(check_condition);

 /* put command complete into the message in buffer */
 buffer_table[msg_in_buf].address = getPhysAddr(command_complete);
 return_flag = 2; /* we sent a check condition */

 /* put the unit attention sense data in the sense data buffer */
 buffer_table[sense_data_buf].address = getPhysAddr(reset_sense_data);

 /* return needed variables to their initial values */
 first = 1;
 done_synch = 0;
 done_wide = 0;
}

static void
init_irq()
{
 disable(); /* disable interrupts */

 /* switch on the interrupt assigned by the PCI BIOS */
 switch(device[selected_device].intl)
  {
	case 9:  setvect(IRQ9, isr); /* set irq interrupt vector */
				old_irq_mask = inportb(PIC2_MASK_REG); /* get old int mask */

				/* write the new interrupt mask into the PIC mask register */
				outportb(PIC2_MASK_REG, IRQ9_mask & old_irq_mask);
				break;

	case 10:	setvect(IRQ10, isr); /* set irq interrupt vector */
				old_irq_mask = inportb(PIC2_MASK_REG); /* get old int mask */

				/* write the new interrupt mask into the PIC mask register */
				outportb(PIC2_MASK_REG, IRQ10_mask & old_irq_mask);
				break;

	case 11: setvect(IRQ11, isr); /* set irq interrupt vector */
				old_irq_mask = inportb(PIC2_MASK_REG); /* get old int mask */

				/* write the new interrupt mask into the PIC mask register */
				outportb(PIC2_MASK_REG, IRQ11_mask & old_irq_mask);
				break;

	case 12: setvect(IRQ12, isr); /* set irq interrupt vector */
				old_irq_mask = inportb(PIC2_MASK_REG); /* get old int mask */

				/* write the new interrupt mask into the PIC mask register */
				outportb(PIC2_MASK_REG, IRQ12_mask & old_irq_mask);
				break;

	case 13: setvect(IRQ13, isr); /* set irq interrupt vector */
				old_irq_mask = inportb(PIC2_MASK_REG); /* get old int mask */

				/* write the new interrupt mask into the PIC mask register */
				outportb(PIC2_MASK_REG, IRQ13_mask & old_irq_mask);
				break;

	case 14: setvect(IRQ14, isr); /* set irq interrupt vector */
				old_irq_mask = inportb(PIC2_MASK_REG); /* get old int mask */

				/* write the new interrupt mask into the PIC mask register */
				outportb(PIC2_MASK_REG, IRQ14_mask & old_irq_mask);
				break;

	case 15: setvect(IRQ15, isr); /* set irq interrupt vector */
				old_irq_mask = inportb(PIC2_MASK_REG); /* get old int mask */

				/* write the new interrupt mask into the PIC mask register */
				outportb(PIC2_MASK_REG, IRQ15_mask & old_irq_mask);
				break;
  }
 enable(); /* enable interrupts */
}

static void
release_irq()
{
 disable();	/* disable interrupts */
 outportb(PIC2_MASK_REG, old_irq_mask); /* restore the old mask */
 enable(); /* enable interrupts */
}

