/********************************************************/
/* PCI.C    Host Platform dependant functions.          */
/*      This version for PCI machines.                  */
/********************************************************/
/*      Support ONLY for 2.0 'cuz Function ID changed   */
/*                                         09/29/93 RLS */
/*      Added 2.0 BIOS support, and multiple device     */
/*      scanning:                          09/28/93 RLS */
/*      Removed fake PCI stub BIOS support 07/21/93 RLS */
/*		  Cleaned up code and made it generic 08/12/95 ACB */
/********************************************************/

#include "pci.h"
#include <dos.h>
#include <stdio.h>

/* Required for inline code to run through TASM to access 386   */
/* registers used by PCI BIOS 2.0+                              */
#pragma inline

/* Function prototypes */
int PCI_FindSYMDevice(device_record *Device);
void PCI_SetConfigRegister(device_record *Device, unsigned int offset,
							  unsigned long value);
unsigned long PCI_GetConfigRegister(device_record *Device, unsigned int offset);
unsigned int  PCI_GetPCIBIOSVersion(pci_bios_info *pcibios);

/******************************************************************************/
/* Function: PCI_FindSYMDevice
	Purpose : To determine if a given Symbios Logic device is installed in
				 the PCI system.
	Input: Structure device_record with the following entries filled in:
					device_id	The device id being searched for
					device_index	The index of the device being searched for
	Output:	This function fills in the rest of the Device data structure
				if a device is found and returns PCI_DEVICE_FOUND.
				If no device was found with that ID and INDEX then
				PCI_NO_DEVICE_FOUND is returned with the Device structure
				unmodified.
	Assumptions: None
	Other functions called: PCI_GetPCIBIOSVersion to make sure there's a PCI BIOS
									PCI_GetConfigRegister to fill in the device rev id
*/
/******************************************************************************/

int PCI_FindSYMDevice(device_record *Device)
{

	struct REGPACK regs;			/* register structure required by INTR function */
	unsigned long temp_config;	/* temporary holding place for config dword */

	/* make sure there is a PCI bios present */
	if (PCI_GetPCIBIOSVersion(NULL) == PCI_NO_BIOS) {
		return(PCI_NO_DEVICE_FOUND);
	}

	/* set up registers for interrupt call */
	regs.r_ax = ((PCI_FUNCTION_ID << 8) | (PCI_FIND_DEVICE));
	regs.r_cx = Device->device_id;
	regs.r_dx = PCI_SYM_VENDOR_ID;
	regs.r_si = Device->device_index;
	intr(PCI_BIOS_INT,&regs);

	if (regs.r_flags & 0x01) {		/* Carry set, No such device found */
			return(PCI_NO_DEVICE_FOUND);
	}

	/* fill in device structure */
	Device->bus_num  =  (regs.r_bx & 0xFF00) >> 8; /* BH = bus num */
	Device->device_num =  (regs.r_bx & 0x00FF);	  /* BL = device num/function */
	Device->function = (regs.r_bx & 0x0007);		  /* BL = function */

	/* Command register */
	Device->command = (unsigned int) PCI_GetConfigRegister(Device,
																			PCI_CONFIG_REG_CMD);
	/* Revision ID */
	Device->rev_id =  (unsigned char) PCI_GetConfigRegister(Device,
																			PCI_CONFIG_REG_REVID);
	/* Cach Line Size and Latency Timer */
	temp_config = PCI_GetConfigRegister(Device, PCI_CONFIG_REG_CLS);
	Device->CLS = (unsigned char) temp_config;
	Device->latency = (unsigned char) (temp_config >> 8);

	/* Cardbus CIS ptr */
	Device->CIS = PCI_GetConfigRegister(Device, PCI_CONFIG_REG_CIS);

	/* Subsystem ID and Subsystem Vendor ID */
	temp_config = PCI_GetConfigRegister(Device, PCI_CONFIG_REG_SUBV);
	Device->sub_ven_id = (unsigned int) temp_config;
	Device->sub_id = (unsigned int) (temp_config >> 16);

	/* ROM Base addr */
	Device->ROM_base_addr = PCI_GetConfigRegister(Device, PCI_CONFIG_REG_ROM);

	/* Int line and Int pin */
	temp_config = PCI_GetConfigRegister(Device, PCI_CONFIG_REG_INTL);
	Device->intl = (unsigned char) temp_config;
	Device->intp = (unsigned char) (temp_config >> 8);


	return(PCI_DEVICE_FOUND);
}



/******************************************************************************/
/* Function: PCI_SetConfigRegister
	Purpose: To write a particular devices configuration register
	Input: device_record data structure indicating the device to write to
			 offset - the dword offset in config space to write
			 value - the value to write to the config register
	Output: Nothing
	Assumptions: None
	Restrictions: Only dword writes to config space are allowed
	Other functions called: PCI_GetPCIBIOSVersion to make sure there's a PCI BIOS
*/
/******************************************************************************/

void PCI_SetConfigRegister(device_record *Device, unsigned int offset,
							  unsigned long value)
{

	/* register variables */
	unsigned int r_ax, r_bx, r_di, r_dx, r_cx;
	unsigned long r_ecx;

	unsigned int pci_version; 		/* used to store pci bios version */

	r_ax = ((PCI_FUNCTION_ID << 8) | (PCI_WRITE_CONFIG_DWORD));
	r_bx = ((Device->bus_num & 0xff) << 8) | (Device->device_num & 0xff);
	r_di = offset;

	pci_version = PCI_GetPCIBIOSVersion(NULL);

	if (pci_version == PCI_BIOS_REV_1X) {       /* Version 1.x  */
		r_dx = (unsigned int) (value >> 16);
		r_cx = (unsigned int) value;
	}
	else if (pci_version == PCI_BIOS_REV_2X) { /* Version 2.x  */
		r_ecx = value;
	}
	else {		/* oh no!!!!!, no bios */
		return;
	}

	/* do PCI function call */
	asm     {
		.386
		mov     ax,[r_ax]
		mov     bx,[r_bx]
		mov     cx,[r_cx]
		mov     dx,[r_dx]
		mov     ecx,DWORD PTR [r_ecx]
		mov     di,[r_di]
		int     PCI_BIOS_INT
	}
}



/******************************************************************************/
/* Function: PCI_GetConfigRegister
	Purpose: To read a particular devices configuration register
	Input: device_record data structure indicating the device to read from
			 offset - the dword offset in config space to read
	Output: Dword config register value
	Assumptions: None
	Restrictions: Only dword reads to config space are allowed
	Other functions called: PCI_GetPCIBIOSVersion to make sure there's a PCI BIOS
*/
/******************************************************************************/

unsigned long PCI_GetConfigRegister(device_record *Device, unsigned int offset)
{

	/* register variables */
	unsigned int r_ax, r_bx, r_di, r_dx, r_cx;
	unsigned long r_ecx;

	unsigned int pci_version; 	/* used to store pci bios version */
	unsigned long result=0;		/* result of PCI Config Read */

	pci_version = PCI_GetPCIBIOSVersion(NULL);

	/* make sure there is a PCI bios present */
	if (pci_version == PCI_NO_BIOS) {
		return(0xFFFFFFFFL);
	}

	r_ax = ((PCI_FUNCTION_ID << 8) | (PCI_READ_CONFIG_DWORD));
	r_bx = ((Device->bus_num & 0xff) << 8) | (Device->device_num & 0xff);
	r_di = offset;

	asm     {
		.386
		mov     ax,[r_ax]
		mov     bx,[r_bx]
		mov     di,[r_di]
		int     PCI_BIOS_INT
		mov     [r_dx],dx
		mov     [r_cx],cx
		mov     DWORD PTR [r_ecx],ecx
	}

	if (pci_version == PCI_BIOS_REV_1X) {       /* Version 1.x  */
		result = r_dx;
		result = (result << 16) | r_cx;
	}
	else if (pci_version == PCI_BIOS_REV_2X)  { /* Version 2.x  */
		result = r_ecx;
	}
	else {		/* oh no!!!!!, unknown pci bios version */
		result = 0xFFFFFFFFL;
	}

	return(result);
}

/******************************************************************************/
/* Function: PCI_GetPCIBIOSVersion
	Purpose: To determine what (if any) PCI BIOS version is out there
	Input: pointer to pci_bios_info data structure
			 If this pointer is null then no attempt is made to fill in the
			 structure (can be used for just assuring the existance of a PCI BIOS),
			 if it is a valid pointer then info on the PCI BIOS is returned.
			 This includes the Interface Version, Access Mechanism's supported
			 and the last pci bus in the system.
	Output: PCI_BIOS_REV_1X	- PCI BIOS version 1.x found
			  PCI_BIOS_REV_2X - PCI BIOS version 2.x found
			  PCI_UNKNOWN_BIOS - PCI BIOS found but unknown version
			  PCI_NO_BIOS - No PCI BIOS present in the system
			  If pcibios != NULL then BIOS info is filled in and returned in the
			  structure.
	Assumptions: None
	Restrictions: None
	Other functions called: None
*/
/******************************************************************************/

unsigned int PCI_GetPCIBIOSVersion(pci_bios_info *pcibios)
{
	/* register variables */
	unsigned int r_ax, r_bx, r_cx, r_dx;
	unsigned long r_edx;

	unsigned long sig;		/* PCI Bios Signature string */

	r_ax = ((PCI_FUNCTION_ID << 8) | PCI_BIOS_PRESENT);

	asm     {
		.386
		mov     ax,[r_ax]
		int     PCI_BIOS_INT
		mov     DWORD PTR [r_edx],edx
		mov     [r_dx],dx
		mov     [r_cx],cx
		mov     [r_bx],bx
		mov	  [r_ax],ax
	}

	/* make sig = " ICP" */
	sig = ('I' | (' ' << 8));
	sig = ((sig << 16) | ('P' | ('C' << 8)));

	if (r_dx == (('P') | ('C' << 8)))
	{
		if (((r_cx & 0xff) == 'I') && ((r_bx & 0xff00) == 0x0100)) {
			/* Rev 1.x BIOS       */
			if (pcibios != NULL) {
				pcibios->access_mech = r_ax & 0xFF;		/* AL = access mech */
				pcibios->interface_lvl = r_bx;			/* BX = interface level */
				pcibios->last_bus = r_cx & 0xFF;			/* CL = last pci bus */
			}
			return(PCI_BIOS_REV_1X);
		}
		else if (r_edx == sig) {
			/* Rev 2.x BIOS       */
			if (pcibios != NULL) {
				pcibios->access_mech = r_ax & 0xFF;		/* AL = access mech */
				pcibios->interface_lvl = r_bx;			/* BX = interface level */
				pcibios->last_bus = r_cx & 0xFF;			/* CL = last pci bus */
			}
			return(PCI_BIOS_REV_2X);
		}
		else {
			return(PCI_UNKNOWN_BIOS);
		}
	}
	else {
		return(PCI_NO_BIOS);
	}
}
